package de.hs_mannheim.informatik.spreadsheet;

import org.junit.Test;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A simplified spreadsheet class for the PR1 programming lab at Hochschule Mannheim.
 * One aspect worth mentioning is that it only supports long numbers, not doubles.
 *
 * @author  Oliver Hummel
 */
public class Spreadsheet {
	Cell[][] cells;

	/**
	 * Constructor that creates a Spreadsheet of size rows * cols.
	 * @param rows number of rows
	 * @param cols number of columns
	 */
	public Spreadsheet(int rows, int cols) {

		// TODO limit the maximum size on 99 (1..99) rows and 26 (A..Z) columns

		cells = new Cell[rows][cols];

		for (int r = 0; r < rows; r++)
			for (int c = 0; c < cols; c++)
				cells[r][c] = new Cell();
		//Fehlermeldung wenn es über 99 Zeilen und 26 Spalten sind
		//Oben genanntes bringt nichts. Bugfix durch zurücksetzen auf vorherigen Wert. Fehlermeldung unnötig durch if Abfrage.
	}

	// -----
	// retrieve or change values of cells

	private String get(int row, int col) {
		return cells[row][col].getValue();
	}

	public String get(String cellName) {
		cellName = cellName.toUpperCase();
		return get(getRow(cellName), getCol(cellName));
	}

	private void put(int row, int col, String value) {
		if (!value.startsWith("="))
			cells[row][col].setValue(value);
		else {
			cells[row][col].setFormula(value);
			evaluateCell(row, col);
		}
	}

	public void put(String cellName, String value) {
		cellName = cellName.toUpperCase();
		put(getRow(cellName), getCol(cellName), value);
	}

	private int getCol(String cellName) {
		return cellName.charAt(0) - 'A';
	}

	private int getRow(String cellName) {
		int ergebnis = 0;
		for (int i = 1; i < cellName.length() ; i++) {
			ergebnis = (ergebnis * 10) + cellName.charAt(i) - '0';


		}
		return ergebnis-1;

	}

	// -----
	// business logic

	/**
	 * A method for reading in data from a CSV file.
	 * @param path The file to read.
	 * @param separator The char used to split up the input, e.g. a comma or a semicolon.
	 * @param startCellName The upper left cell where data from the CSV file should be inserted.
	 * @return Nothing.
	 * @exception IOException If path does not exist.
	 */
	public void readCsv(String path, char separator, String startCellName) throws FileNotFoundException {
		// TODO: implement this
		ArrayList<String> lines = readFile(path);
		String[] seperateLines = new String[lines.size()];
		int startingCellRow = getRow(startCellName);
		int startingCellCol = getCol(startCellName);
		String[][] quickSave = new String[cells.length - startingCellRow][cells[0].length - startingCellCol];
		for (int i = 0; i < seperateLines.length; i++) {
			quickSave[i] = lines.get(i).split(separator+""); //Seperator erstellt
			for (int j = 0; j < quickSave[i].length; j++) {
				put(startingCellRow+i,startingCellCol+j,quickSave[i][j]);
			}
		}
	}

	/**
	 *
	 * @param path Dateipfad der Datei die von der Methode ausgelesen werden soll
	 * @return lines
	 * @throws FileNotFoundException
	 */
	public static ArrayList<String> readFile(String path) throws FileNotFoundException {
		ArrayList<String> lines = new ArrayList<>();
		Scanner sc = new Scanner(new File(path));

		while (sc.hasNextLine()) {
			lines.add(sc.nextLine());
		}

		sc.close();

		return lines;
	}




	/**
	 * A method for saving data to a CSV file.
	 * @param path The file to write.
	 * @return Nothing.
	 * @exception IOException If path does not exist.
	 */
	public void saveCsv(String path) throws FileNotFoundException {
		PrintWriter out = new PrintWriter(path);

		for (Cell[] row : cells) {
			for (Cell cell : row) {
				if (!cell.getFormula().isEmpty())
					out.print("=" + cell.getFormula());
				else
					out.print(cell.getValue());

				if (cell != row[cells[0].length-1])
					out.print(",");
			}
			out.println();
		}

		out.close();
	}

	/**
	 * Methode welche die Eingabe überprüft und anhand dessen, dann auf abgebildete Methoden zugreift
	 * und das Ergebnis wiedergibt
	 * @param row steht für die Zeile
	 * @param col steht für die Spalte
	 */
	private void evaluateCell(int row, int col) {
		String formula = cells[row][col].getFormula();
		String result = "";

		if (formula.startsWith("SUMME(")) // e.g. SUMME(A3:A8)
			result = "" + sum(formula.substring(6, 8), formula.substring(9, 11));  // TODO adapt to cells with two digits
		else if (formula.startsWith("PRODUKT("))  // e.g. PRODUKT(A3:B9)
			result = "" + multiplier(formula.substring(8, 10), formula.substring(11, 13)); // TODO
		else if (formula.startsWith("MITTELWERT("))  // e.g. MITTELWERT(A3:A5)
			result = "" + mittelwert(formula.substring(11, 13), formula.substring(14, 16)); // TODO
		else if (formula.startsWith("STABW("))  // e.g. STABW(C6:D8) -> Standardabweichung
			result = "TODO"; // TODO
		else if (formula.startsWith("MIN("))  // e.g. MIN(C13:H13) -> größter Wert
			result = "" + min(formula.substring(4, 6), formula.substring(7, 9)); // TODO
		else if (formula.startsWith("MAX("))  // e.g. MAX(A1:A10) -> Standardabweichung
			result = "" + max(formula.substring(4, 6), formula.substring(7, 9)); // TODO
		else if (!formula.isEmpty()) {
			try {
				result = "" + calculate(formula);
			} catch(ArithmeticException ae) {
				result = "exc.";
			}
		}

		cells[row][col].setValue("" + result);
	}
	//substring start und ende auf die entsprechende Wortlänge bei der Bedingung anpassen
	/**
	 * Method for calculating the sum of a rectangular block of cells, such as from A1 to B3.
	 * @param startCellName The name of the cell in the upper left corner of the rectangle.
	 * @param endCellName The name of the cell in the lower right corner of the rectangle.
	 * @return The sum calculated.
	 */
	private long sum(String startCellName, String endCellName) {

		int startRow = getRow(startCellName);
		int startCol = getCol(startCellName);
		int endRow = getRow(endCellName);
		int endCol = getCol(endCellName);

		long result = 0;

		for (int i = startRow; i <= endRow; i++) {
			for (int j = startCol; j <= endCol; j++) {
				result += (Integer.parseInt(cells[i][j].getValue()));
			}
		}

		return result;
	}

	/**
	 * Methode zum berechnen des Produkts eines viereckigen Blocks aus Zellen. Beispiel A2 bis B2
	 * @param startCellName  The name of the cell in the upper left corner of the rectangle.
	 * @param endCellName The name of the cell in the lower right corner of the rectangle.
	 * @return das Produkt. Hier result
	 */
	private long multiplier(String startCellName, String endCellName) {

		int startRow = getRow(startCellName);
		int startCol = getCol(startCellName);
		int endRow = getRow(endCellName);
		int endCol = getCol(endCellName);

		long result = 1; // 1 da hier Multiplikation startet

		for (int i = startRow; i <= endRow; i++) {
			for (int j = startCol; j <= endCol; j++) {
				result *= (Integer.parseInt(cells[i][j].getValue()));
			}
		}

		return result;
	}
	/**
	 * Methode zum berechnen des Mittelwerts eines viereckigen Blocks aus Zellen. Beispiel A2 bis B2
	 * @param startCellName  The name of the cell in the upper left corner of the rectangle.
	 * @param endCellName The name of the cell in the lower right corner of the rectangle.
	 * @return der Mittelwert. Hier result
	 */
	private long mittelwert(String startCellName, String endCellName) {

		int startRow = getRow(startCellName);
		int startCol = getCol(startCellName);
		int endRow = getRow(endCellName);
		int endCol = getCol(endCellName);
		int counter = 0;
		long summe = 0;

		long result = 0;

		for (int i = startRow; i <= endRow; i++) {
			for (int j = startCol; j <= endCol; j++) {
				counter++;
				summe += (Integer.parseInt(cells[i][j].getValue()));
			}
		}
		result = summe / counter;

		return result;
	}
	/**
	 * Methode zum berechnen des Minimums eines viereckigen Blocks aus Zellen. Beispiel A2 bis B2
	 * @param startCellName  The name of the cell in the upper left corner of the rectangle.
	 * @param endCellName The name of the cell in the lower right corner of the rectangle.
	 * @return die kleinste Zahl. Hier result
	 */
	 private long min(String startCellName, String endCellName) {
		 int startRow = getRow(startCellName);
		 int startCol = getCol(startCellName);
		 int endRow = getRow(endCellName);
		 int endCol = getCol(endCellName);
		 long result = 0;

		 for (int i = startRow; i <= endRow; i++) {
			 for (int j = startCol; j <= endCol; j++) {

				 result = (Integer.parseInt(cells[i][j].getValue()));
				 if(result >  (Integer.parseInt(cells[i][j].getValue())))  //größer result für min Wert
				 {
					 result = (Integer.parseInt(cells[i][j].getValue()));
				 }

			 }
		 }
		 return result;
	 }
	/**
	 * Methode zum berechnen des Maximums eines viereckigen Blocks aus Zellen. Beispiel A2 bis B2
	 * @param startCellName  The name of the cell in the upper left corner of the rectangle.
	 * @param endCellName The name of the cell in the lower right corner of the rectangle.
	 * @return das Maximum. Hier result
	 */
	private long max(String startCellName, String endCellName) {
		int startRow = getRow(startCellName);
		int startCol = getCol(startCellName);
		int endRow = getRow(endCellName);
		int endCol = getCol(endCellName);
		long result = 0;

		for (int i = startRow; i <= endRow; i++) {
			for (int j = startCol; j <= endCol; j++) {

				if(result <  (Integer.parseInt(cells[i][j].getValue()))) // kleiner result für max Wert
				{
					result = (Integer.parseInt(cells[i][j].getValue()));
				}

			}
		}
		return result;
	}


		// TODO implement


	/**
	 * This method calculates the result of a "normal" algebraic expression. It only needs to support
	 * expressions like =B4 or =2+A3-B2, i.e. only with int numbers and other cells and with plus,
	 * minus, times, split only. An expression always starts with either a number or a cell name. If it 
	 * continues, it is guaranteed that this is followed by an operator and either a number or a 
	 * cell name again. It is NOT required to implement dot before dash or parentheses in formulas.
	 * @param formula The expression to be evaluated.
	 * @return The result calculated.
	 */

	public long calculate(String formula) throws ArithmeticException {
		Matcher m = Pattern.compile("([A-Z][0-9]*)|[-\\+\\*/]|[0-9]*").matcher(formula);

		long result = 0;
		int zwischenspeicher = 0;
		int counter = 0;
		int newNumber = 0;

		// TODO implement


		while (m.find()) {        // m.find() must always be used before m.group()
			String s = m.group();
			if (!s.isEmpty()) {
				//System.out.println(s);
				counter++;

				switch (counter){
					case 1:
						try {
							result = Integer.parseInt(s);
						}catch (Exception ignored){
							result = Integer.parseInt(get(s));
						}
						break;
					case 2:
						switch(s) {
							case "+":
								zwischenspeicher = 1;
								break;
							case "-":
								zwischenspeicher = 2;
								break;
							case "*":
								zwischenspeicher = 3;
								break;
							case "/":
								zwischenspeicher = 4;
								break;
							default:
								System.out.println("ERROR");
								break;
						}
						break;
					case 3:
						try {
							newNumber = Integer.parseInt(s);
						}catch (Exception ignored){
							newNumber = Integer.parseInt(get(s));
						}
						result = rechner(result, zwischenspeicher, newNumber);

						counter = 1;
						break;
					default:
						System.out.println("ERROR");
						break;
				}
			}
		}

		return result;
	}

	/**
	 *
	 * @param a steht für die Zahl "a"
	 * @param auswahl steht für die entsprechende Rechnung
	 * @param b	steht für die Zahl "b"
	 * @return die Rechnung
	 */
	public static long rechner(long a, int auswahl, int b) {
		switch(auswahl){
			case 1:
				return a + b;
			case 2:
				return a - b;
			case 3:
				return a * b;
			case 4:
				return a / b;
			default:
				System.out.println("ERROR");
				break;
		}
		return 0;
	}
	// -----

	public String toString() {
		StringBuilder sb = new StringBuilder();

		sb.append("    ");
		for (int i = 0; i < cells[0].length; i++) {
			sb.append("  " + (char)('A'+ i) + "  | ");
		}

		int rc = 1;
		for (Cell[] r : cells) {
			sb.append(System.lineSeparator());
			sb.append(String.format("%2s", rc++) + ": ");

			for (Cell c : r) {
				sb.append(c + " | ");
			}

		}
		return sb.toString();
	}

}